function [U, V] = aca(varargin)
%ACA Adaptive cross approximation with partial pivoting
%
%    [U, V] = ACA(A, TOL) uses Adaptive Cross Approximation to 
%    determine a low-rank approximation of the matrix A. TOL is
%    the relative tolerance used for truncation. 
%
%    [U, V] = ACA(AFUN, M, N, TOL) uses the function handle AFUN to
%    evaluate the entries of A, by calling AFUN(i, j). The integers M and
%    N are the number of rows and columns of the matrix. 
%
%    [U, V] = ACA(A, TOL, COMP, DEBUG) optionally allows to enable/disable
%    recompression at the end (COMP), and enable/disable the debug (DEBUG).
%
% Authors: Stefano Massei <stefano.massei@epfl.ch>
%          Leonardo Robol <leonardo.robol@unipi.it>

Afun = varargin{1};

debug = false;
comp  = false;

if nargin >= 2 && nargin <= 3
	m = size(Afun, 1);
	n = size(Afun, 2);
	tol = varargin{2};
	
	if nargin == 3
		comp = varargin{3};
	end	
elseif nargin >= 4
	tol = varargin{4};
	
	% We could have two cases: either varargin{1} is a handle function, and
	% then the dimensions are specified in varargin{2:3}, or it is a
	% matrix, and then the m, n are dropped
	if ~isa(Afun, 'function_handle')
		if nargin > 4
			error('Too many arguments specified');
		end
		
		comp = varargin{3};
		debug = varargin{4};
	else	
		m = varargin{2};
		n = varargin{3};
	
		if nargin >= 5
			comp = varargin{5};
		end
	
		if nargin >= 6
			debug = varargin{6};
		end
	end
else
	error('Incorrect number of arguments');
end

U = zeros(m, 0);
V = zeros(n, 0);
k = 1;

% Select the first pivot -- some randomization going on here. 
first_indices = unique(randi(m, 4, 1));
rows = Afun(first_indices, 1:n);

[~, ind] = max(max(abs(rows), [], 2));

while k < min(m,n)
	b = Afun(ind, 1:n) - U(ind, :) * V';
	[~, new_ind] = max(abs(b));
	if debug
		fprintf('Iteration: %d Pivot at (%d,%d): %e\n', k, ind, new_ind, b(new_ind))
	end
	a = Afun((1:m)', new_ind) - U * V(new_ind, :)';
	a = a/b(new_ind);
	U = [U, a];
	V = [V, b'];
	[~, tind] = max(abs(a([1:ind - 1, ind + 1:m])));  
    if tind >= ind
		ind = tind + 1;
	else
		ind = tind;
	end

	if k == 1
		nrm = norm(a) * norm(b);
	end
	k = k + 1;
	if norm(a) * norm(b) < tol * nrm
		break
	end
end

if comp
	[QU, RU] = qr(U, 0); % Re-compression
	[QV, RV] = qr(V, 0);
	[U, S, V] = svd(RU * RV', 'econ');
	rk = sum(diag(S) > tol * S(1,1));
	U = QU * U(:,1:rk) * sqrt(S(1:rk,1:rk));
    V = QV * V(:,1:rk) * sqrt(S(1:rk,1:rk));
end

end

