% 
% Solution of exercise 3 for the lab session on matrix equations and
% hierarchical matrices. 
%

% Some parameters:
n = 2^10; % Number of discretization points on the edge of the grid. 
nt = 100; % Number of time steps to perform. 

% Problem statement: we would like to solve a 2D PDE on [0, 1]^2, given by 
%    du/dt = \Delta u + f 
% subject to Dirichlet zero boundary conditions. We assume f is given.
% Instead of solving for the steady-state as we have done in the previous
% exercise, we now integrate over the time interval [0, T], and assuming to
% have a diffusion coefficient of D. 

% Both the diffusion and the source term are allowed to change over time.
% The following are just possible examples, you may try to change them to
% suit your liking. 

% Maximum time
T = 10;

% Diffusion coefficient
D = @(t) 0.1 + 0.01 * t;

% f(x,y), source function (time independent)
f = @(x,y,t) exp(- x.^2 - y.^2) .* (T - t/2);

% Construct the 1D discretization of the Laplacian. 
h = 1 ./ (n - 1);
x = linspace(0, 1, n);

% The system has size (n-2) x (n-2) because we have imposed the Dirichlet
% boundary conditions.
A = h^(-2) * spdiags(ones(n-2, 1) * [1 -2 1], -1:1, n - 2, n - 2);

% We can now solve the matrix equation AX + XA + F = 0, using the
% factorization F = UF * VF';
UF = g1(x(2:end-1)).'; VF = g2(x(2:end-1)).';

% Time discretization
t = linspace(0, T, nt);
dt = T ./ (nt - 1);

% Cell-array where we store the solution at each time-step
UX = cell(1, nt);
VX = cell(1, nt);

% Initialize u = 0 at the beginning
UX{1} = zeros(n-2,0);
VX{1} = zeros(n-2,0);

for j = 2 : nt
	[UF, VF] = aca(@(i,s) f(x(s), x(i).', t(j)), n, n, 1e-6);
	
	UF = UF(2:end-1,:); VF = VF(2:end-1,:);
	
	% Compute the RHS, and recompress it if needed
	UR = - [ dt * UF, UX{j-1} ];
	VR = [ VF, VX{j-1} ];
	
	[UR, VR] = recompress_low_rank(UR, VR, 1e-6);	

	% System to solve for the time-stepping scheme. Notice that we are
	% splitting the identity matrix half for each factor, to keep the problem
	% symmetric. 
	AA = .5 * speye(n-2) - dt * D(t(j)) * A;
	
	% Compute the solution at the next time step
	[UX{j}, VX{j}] = ek_sylv(AA, AA, UR, VR, inf, 1e-6, false, 'fro');
end

% Find out the right scale for the plot, trying to estimate a box that
% always contain the solution
maxsol = 0;
for j = 1 : nt
	maxsol = max(maxsol, norm(UX{j}, inf) * norm(VX{j}, inf));
end

% Plot the solution, if it's not too large, otherwise we subsample it to k
% points where k is small enough
timer = tic;
for j = 1 : nt
	
	% Try to make the animation real time, is possible
	timetowait = max(0, t(j) - toc(timer));
	pause(timetowait);
	
	if n <= 256
		mesh(x, x, blkdiag(0, UX{j} * VX{j}', 0));
	else	
		k = 256;
		indices = ceil( (1:k) ./ k * size(UX{j}, 1) );

		% Include boundary conditions for the indices of x. 
		allindices = [ 1, indices, n ];

		% Add the zero Dirichlet boundary conditions
		X = blkdiag(0, UX{j}(indices, :) * VX{j}(indices, :)', 0);

		% Plot the subsampled mesh
		mesh(x(allindices), x(allindices), X);
	end
	
	% We make sure that the axis stays the same -- so that we do not see
	% the animation bumping up and down
	axis([0 1 0 1 -maxsol maxsol]);
	drawnow;
end

function [U, V] = recompress_low_rank(U, V, tol)
	[QU, RU] = qr(U, 0);
	[QV, RV] = qr(V, 0);
	
	[U, S, V] = svd(RU * RV');
	rk = sum(diag(S) > S(1,1) * tol);
	
	U = QU * ( U(:,1:rk) * sqrt(S(1:rk,1:rk)) );
	V = QV * ( V(:,1:rk) * sqrt(S(1:rk,1:rk)) );
end